local function to_color_component(str)
  local value = tonumber(str)
  if value and value >= 0 and value <= 255 then
    return value
  end
end

------------------------------------------------------------------------------------------------

AbstractEditor = Widget:derive()

function AbstractEditor:adjust_value(func)
  local entity = Entity.from_data(self.ent_data)
  local ent_field = self.ent_field
  entity[ent_field] = func(entity[ent_field])
  entity:on_field_changed(ent_field)
end

function AbstractEditor:load_value(value)
end

function AbstractEditor:save_value(value)
  local entity = Entity.from_data(self.ent_data)
  local ent_field = self.ent_field
  entity[ent_field] = value
  entity:on_field_changed(ent_field)
end

------------------------------------------------------------------------------------------------

BoolEditor = AbstractEditor:derive("bool-editor")

function BoolEditor:create(parent)
  self.check_box = self:add_element(parent, {
    type = "checkbox",
    name = "editor",
    state = false,
    style = "tux_editor_bool_box",
  })
end


function BoolEditor:load_value(value)
  self.check_box.state = value
end

function BoolEditor:on_changed(event)
  self:save_value(self.check_box.state)
end

------------------------------------------------------------------------------------------------

ColorEditor = AbstractEditor:derive("color-editor")

function ColorEditor:create(parent)
  self.red_field = self:add_element(parent, {
    type = "textfield",
    name = "editor_r",
    style = "tux_editor_color_field",
    text = "255",
  })

  self.green_field = self:add_element(parent, {
    type = "textfield",
    name = "editor_g",
    style = "tux_editor_color_field",
    text = "255",
  })

  self.blue_field = self:add_element(parent, {
    type = "textfield",
    name = "editor_b",
    style = "tux_editor_color_field",
    text = "255",
  })
end

function ColorEditor:load_component(element, value)
  local string_value = tostring(value)
  if element.text ~= string_value then
    element.text = string_value
  end 
end

function ColorEditor:load_value(color)
  self:load_component(self.red_field, color.r)
  self:load_component(self.green_field, color.g)
  self:load_component(self.blue_field, color.b)
end

function ColorEditor:on_changed(event)
  local r = self:validate_color_field(self.red_field)
  local g = self:validate_color_field(self.green_field)
  local b = self:validate_color_field(self.blue_field)
  if r and g and b then
    self:save_value{r = r, g = g, b = b}
  end
end

function ColorEditor:validate_color_field(element)
  local value = to_color_component(element.text)

  if value then
    element.style = "tux_editor_color_field"
    return value
  else
    element.style = "tux_editor_color_field_red"
  end
end

------------------------------------------------------------------------------------------------

ComplexVariantEditor = AbstractEditor:derive("complex-variant-editor")
ComplexVariantEditor.variants = {}

function ComplexVariantEditor:create(parent)
  local items = {}

  for _, entry in pairs(self.variants) do
    table.insert(items, entry.caption)
  end

  self.dropdown = self:add_element(parent, {
    type = "drop-down",
    name = "editor",
    style = "tux_editor_complex_variant",
    items = items,
    selected_index = 1,
  })

  self.button = self:add_element(parent, {
    type = "button",
    name = "edit-variant",
    style = "tux_editor_complex_variant_button",
  })
end

function ComplexVariantEditor:create_config_frame()
  local index = self.dropdown.selected_index
  local editor_class = self.variants[index].editor

  if editor_class then
    local player_id = self.button.player_index
    local frame_stack = FrameStack:from_player_id(player_id)
    local config_frame = editor_class:new{ent_data = self.ent_data}
    frame_stack:push_frame(self.parent.index + 1, config_frame)
    self.config_frame = config_frame
  end
end

-- function ComplexVariantEditor:destroy()
-- end

function ComplexVariantEditor:destroy_config_frame()
  self.config_frame:destroy()
  self.config_frame = nil
end

function ComplexVariantEditor:load_value(value)
  for index, entry in ipairs(self.variants) do
    if entry.value == value then
      self.dropdown.selected_index = index
      self:update_button()
      return
    end
  end
end

function ComplexVariantEditor:on_changed(event)
  local index = self.dropdown.selected_index
  local entry = self.variants[index]

  if entry then
    self:save_value(entry.value)  
  end

  self:update_button()

  if self.config_frame and self.config_frame.valid then 
    self:destroy_config_frame()
    self:create_config_frame() 
  end
end

function ComplexVariantEditor:on_destroyed()
  if self.config_frame and self.config_frame.valid then
    self:destroy_config_frame()
  end
end

function ComplexVariantEditor:set_variants(variants)
  self.variants = variants
end

function ComplexVariantEditor:update_button()
  local index = self.dropdown.selected_index
  local entry = self.variants[index]

  if entry and entry.editor then
    self.button.enabled = true
  else
    self.button.enabled = false
  end
end

ComplexVariantEditor:on_clicked("edit-variant", function (self, event)
  if self.config_frame and self.config_frame.valid then
    self:destroy_config_frame()
  else
    self:create_config_frame()
  end
end)

------------------------------------------------------------------------------------------------

IntegerEditor = AbstractEditor:derive("int-editor")

function IntegerEditor:create(parent)
  self.slider = self:add_element(parent, {
    type = "slider",
    name = "editor0",
    style = "tux_editor_integer_slider",
    minimum_value = 0,
    maximum_value = 100000,
  })

  self.value_field = self:add_element(parent, {
    type = "textfield",
    name = "editor1",
    style = "tux_editor_integer_box",
  })
end

function IntegerEditor:load_value(value)
  self.slider.slider_value = value
  self.value_field.text = tostring(value)
end

function IntegerEditor:on_changed(event)
  if event.element == self.slider then
    local new_value = self.slider.slider_value
    local string_value = string.format("%i", new_value)
    self.value_field.text = string_value
    self:save_value(new_value)
  elseif event.element == self.value_field then
    local number = tonumber(self.value_field.text)
    if number then
      self:save_value(number)
    end
  end
end

------------------------------------------------------------------------------------------------

SignalEditor = AbstractEditor:derive("signal-editor")

function SignalEditor:create(parent)
  self.chooser = self:add_element(parent, {
    type = "choose-elem-button",
    name = "editor",
    elem_type = "signal",
    style = "tux_editor_signal",
  })
end

function SignalEditor:load_value(signal)
  self.chooser.elem_value = signal
end

function SignalEditor:on_changed(event)
  self:save_value(self.chooser.elem_value)
end

------------------------------------------------------------------------------------------------

StringEditor = AbstractEditor:derive("string-editor")

function StringEditor:create(parent)
  self.text_field = self:add_element(parent, {
    type = "textfield",
    name = "editor",
    style = "tux_editor_string",
  })
end

function StringEditor:load_value(value)
  self.text_field.text = value
end

function StringEditor:on_changed(event)
  self:save_value(self.text_field.text)
end

------------------------------------------------------------------------------------------------

VariantEditor = AbstractEditor:derive("variant-editor")
VariantEditor.variants = {}

function VariantEditor:create(parent)
  local keys = {}
  
  for _, v in pairs(self.variants) do
    table.insert(keys, v[1])
  end

  self.dropdown = self:add_element(parent, {
    type = "drop-down",
    name = "editor",
    style = "tux_editor_variant",
    items = keys,
  })
end

function VariantEditor:load_value(value)
  for index, entry in ipairs(self.variants) do
    if entry[2] == value then
      self.dropdown.selected_index = index
      return
    end
  end
end

function VariantEditor:on_changed(event)
  local index = self.dropdown.selected_index
  local entry = self.variants[index]
  if entry then
    self:save_value(entry[2])  
  end
end

function VariantEditor:set_variants(variants)
  self.variants = variants
end

------------------------------------------------------------------------------------------------
