local function deep_compare(value0, value1)
  if value0 ~= value1 then
    return false
  end

  local type0 = type(value0)
  local type1 = type(value1)
  if type0 ~= type1 then
    return false
  elseif type0 ~= "table" then
    return false
  end

  for k, v0 in pairs(value0) do
    local v1 = value1[k]
    if not deep_compare(v0, v1) then
      return false
    end
  end

  return true
end

entity_classes = {}

------------------------------------------------------------------------------------------------

Entity = {}

function Entity.global_init()
  global.entity_data = global.entity_data or {}
end

function Entity:derive(name)
  return BaseEntity:derive(name)
end

function Entity.from_data(data)
  if getmetatable(data) then
    return data
  end

  local class = entity_classes[data.type]
  if not class then
    error("warning: discarding invalid entity: " .. data.type)
  end

  setmetatable(data, class.__mt)
  return data
end

function Entity.from_entity(entity)
  local entity_id = entity.unit_number
  local data = global.entity_data[entity_id]
  if data then
    return Entity.from_data(data)
  end
end

------------------------------------------------------------------------------------------------

function Entity.finalize_ghost(ghost)
  local entity_class = entity_classes[ghost.ghost_name]
  if entity_class and entity_class.on_ghost_mined then
    entity_class:on_ghost_mined(ghost)
  end
end

------------------------------------------------------------------------------------------------

function Entity.on_built(event)
  local entity = event.created_entity
  local entity_class = entity_classes[entity.name]
  if not entity_class then
    return
  end

  local entity_id = entity.unit_number
  if entity_id then
    local data = entity_class:new{entity = entity}
    global.entity_data[entity_id] = data
    data:create(event)
  end
end

function Entity.on_died(event) -- convert to ghost?
  local entity = Entity.from_entity(event.entity)
  if entity and entity.on_died then
    local entity_id = event.entity.unit_number
    entity:on_died(event)
    global.entity_data[entity_id] = nil
  end
end

function Entity.on_mined(event)
  local raw_entity = event.entity
  local entity = Entity.from_entity(raw_entity)

  if entity and entity.on_mined then
    local entity_id = raw_entity.unit_number
    entity:on_mined(event)
    global.entity_data[entity_id] = nil
    return
  end

  if raw_entity.name == "entity-ghost" then
    Entity.finalize_ghost(raw_entity)
    return
  end
end

function Entity.on_pre_ghost_deconstructed(event)
  Entity.finalize_ghost(event.ghost)
end

function Entity.on_settings_pasted(event)
  local source = Entity.from_entity(event.source)
  local destination = Entity.from_entity(event.destination)

  if source and destination then
    destination:on_settings_pasted(source)
  end
end

------------------------------------------------------------------------------------------------

BaseEntity = {
  properties = {},
}

function BaseEntity:activate(event)
end

function BaseEntity:add_properties(properties)
  for _, prop in pairs(properties) do
    local name = prop.name
    self[name] = prop.default
    self.properties[name] = prop
  end
end

function BaseEntity:configure(event)
  local config_class = self.config_class
  if config_class then
    local player_id = event.player_index
    local frame_stack = FrameStack:from_player_id(player_id)
    local config = config_class:new{ent_data = self}
    frame_stack:replace_frame(config)
  end
end

function BaseEntity:create(entity)
end

function BaseEntity:derive(name)
  local class = table.deepcopy(self)
  class.__mt = {__index = class}
  class.name = name

  if not name then
    return class
  elseif not entity_classes[name] then
    entity_classes[name] = class
    return class
  else
    error("Entity with this name is already registered")
  end
end

function BaseEntity:extend(extension)
  extension:apply(self)
end

function BaseEntity:new(opts)
  local data = opts or {}
  data.type = self.name
  setmetatable(data, self.__mt)
  return data
end

function BaseEntity:on_field_changed(field)
end

function BaseEntity:on_settings_pasted(source)
  for name, src_prop in pairs(source.properties) do
    local src_active = src_prop.active(source)
    local dst_prop = self.properties[name]

    if src_active and dst_prop then
      local dst_value = self[name]
      local src_value = source[name]
      if not deep_compare(dst_value, src_value) then
        self[name] = table.deepcopy(src_value)
        self:on_field_changed(name)
      end
    end
  end
end

------------------------------------------------------------------------------------------------

script.on_event(defines.events.on_built_entity, Entity.on_built)
script.on_event(defines.events.on_entity_died, Entity.on_died)
script.on_event(defines.events.on_entity_settings_pasted, Entity.on_settings_pasted)
script.on_event(defines.events.on_robot_built_entity, Entity.on_built)
script.on_event(defines.events.on_robot_mined_entity, Entity.on_mined)
script.on_event(defines.events.on_player_mined_entity, Entity.on_mined)
script.on_event(defines.events.on_pre_ghost_deconstructed, Entity.on_pre_ghost_deconstructed)

------------------------------------------------------------------------------------------------

script.on_event("tux-activate-entity", function (event)
	local raw_entity = game.players[event.player_index].selected
  if not raw_entity then 
    return 
  end

  local entity = Entity.from_entity(raw_entity)
  if entity then
    entity:activate(event)
  end
end)

script.on_event("tux-configure-entity", function (event)
	local raw_entity = game.players[event.player_index].selected
  if not raw_entity then 
    return 
  end

  local entity = Entity.from_entity(raw_entity)
  if entity then
    entity:configure(event)
  end
end)

------------------------------------------------------------------------------------------------
