local decoders = {}
local read_any

------------------------------------------------------------------------------------------------

local function byte_u(number, shift)
  return bit32.rshift(number, shift) % 256
end

local function make_const_reader(const)
  return function (data, pos)
    return const, pos
  end
end

local function make_reader(len_reader, rest_reader)
  return function (data, pos)
    local len
    pos, len = len_reader(data, pos)
    return rest_reader(data, pos, len)
  end
end

------------------------------------------------------------------------------------------------

local function read_u8(data, pos)
  return data[pos], pos + 1
end

local function read_u16(data, pos)
  local value = data[pos + 0] * 0x100 + data[pos + 1]
  return value, pos + 2
end

local function read_u32(data, pos)
  local value 
    = data[pos + 0] * 0x1000000 
    + data[pos + 1] * 0x10000 
    + data[pos + 2] * 0x100 
    + data[pos + 3]
  return value, pos + 4
end

local function read_u64(data, pos)
  local value 
    = data[pos + 0] * 0x100000000000000 
    + data[pos + 1] * 0x1000000000000 
    + data[pos + 2] * 0x10000000000
    + data[pos + 3] * 0x100000000
    + data[pos + 4] * 0x1000000
    + data[pos + 5] * 0x10000
    + data[pos + 6] * 0x100
    + data[pos + 7] 
  return value, pos + 8
end

------------------------------------------------------------------------------------------------

local function read_i8(data, pos)
  local value, _ = read_u8(data, pos) 
  if value >= 0x80 then
    value = value - 0x100
  end
  return value, pos + 1
end

local function read_i16(data, pos)
  local value, _ = read_u16(data, pos) 
  if value >= 0x8000 then
    value = value - 0x10000
  end
  return value, pos + 2
end

local function read_i32(data, pos)
  local value, _ = read_u32(data, pos) 
  if value >= 0x80000000 then
    value = value - 0x100000000
  end
  return value, pos + 4
end

------------------------------------------------------------------------------------------------

local function read_f32(data, pos)
  error("not implemented")
end

local function read_f64(data, pos)
  error("not implemented")
end

------------------------------------------------------------------------------------------------

local function read_array(data, pos, len)
  local array, elem = {}
  for i = 1, len do
    elem, pos = read_any(data, pos)
    table.insert(array, elem)
  end
  return array, pos
end

local function read_map(data, pos, len)
  local map, k, v = {}
  for i = 1, len do
    k, pos = read_any(data, pos)
    v, pos = read_any(data, pos)
    map[k] = v
  end
  return map, pos
end

local function read_str(data, pos, len)
  local chars = {}
  for i = 1, len do
    chars[i] = data[pos + i - 1]
  end

  local result = string.char(unpack(chars))
  return result, pos + len
end

------------------------------------------------------------------------------------------------

read_any = function (data, pos)
  local tag = data[pos]
  if tag == nil then
    error("truncated or malformed input stream: pos=" .. pos)
  end

  if tag <= 0x7F then -- fixint
    return tag, pos + 1
  elseif tag <= 0x8F then -- fixmap
    return read_map(data, pos + 1, tag - 0x80)
  elseif tag <= 0x9F then -- fixarray
    return read_array(data, pos + 1, tag - 0x90)
  elseif tag <= 0xBF then -- fixstr
    return read_str(data, pos + 1, tag - 0xA0)
  elseif tag >= 0xE0 then -- fixint
    return tag - 0x100, pos + 1
  end

  local decoder = decoders[tag]
  if decoder then
    return decoder(data, pos + 1)
  else
    error(string.format("no decoder for tag 0x%02x", tag))
  end
end

------------------------------------------------------------------------------------------------

decoders[0xC0] = make_const_reader(nil) -- nil
decoders[0xC1] = function (data, pos) error("not used") end
decoders[0xC2] = make_const_reader(false) -- false
decoders[0xC3] = make_const_reader(true) -- true

decoders[0xC4] = make_reader(read_u8, read_str) -- bin8
decoders[0xC5] = make_reader(read_u16, read_str) -- bin16
decoders[0xC6] = make_reader(read_u32, read_str) -- bin32

decoders[0xCA] = read_f32  -- float32
decoders[0xCB] = read_f64  -- float64
decoders[0xCC] = read_u8  -- uint8
decoders[0xCD] = read_u16 -- uint16
decoders[0xCE] = read_u32 -- uint32
decoders[0xCF] = read_u64 -- uint32
decoders[0xD0] = read_i8  -- int8
decoders[0xD1] = read_i16 -- int16
decoders[0xD2] = read_i32 -- int32
decoders[0xD3] = read_i64 -- int64

decoders[0xD9] = make_reader(read_u8, read_str) -- str8
decoders[0xDA] = make_reader(read_u16, read_str) -- str16
decoders[0xDB] = make_reader(read_u32, read_str) -- str32
decoders[0xDC] = make_reader(read_u16, read_array) -- array16
decoders[0xDD] = make_reader(read_u32, read_array) -- array32
decoders[0xDE] = make_reader(read_u16, read_map) -- map16
decoders[0xDF] = make_reader(read_u32, read_map) -- map32

------------------------------------------------------------------------------------------------

MessagePack = MessagePack or {}

function MessagePack.decode_u8(data_u8)
  local result, _ = read_any(data_u8, 1)
  return result
end

function MessagePack.decode_i32(data_i32)
  local data_u8 = {}

  for _, word in pairs(data_i32) do
    word = word + 0x80000000
    table.insert(data_u8, byte_u(word, 24))
    table.insert(data_u8, byte_u(word, 16))
    table.insert(data_u8, byte_u(word, 8))
    table.insert(data_u8, byte_u(word, 0))
  end

  return MessagePack.decode_u8(data_u8)
end

------------------------------------------------------------------------------------------------
