local encoders = {}
local write = table.insert

------------------------------------------------------------------------------------------------

local function write_any(out, value)
  local value_type = type(value)
  local encoder = encoders[value_type]
  if encoder then
    encoder(value, out)
  else
    error("can't encode " .. value_type)
  end
end

------------------------------------------------------------------------------------------------

local function byte_i(number, shift)
  return bit32.arshift(number, shift) % 256
end

local function byte_u(number, shift)
  return bit32.rshift(number, shift) % 256
end

------------------------------------------------------------------------------------------------

local function write_u8(out, prefix, value)
  write(out, prefix)
  write(out, value)
end

local function write_u16(out, prefix, value)
  write(out, prefix)
  write(out, byte_u(value, 8))
  write(out, byte_u(value, 0))
end

local function write_u32(out, prefix, value)
  write(out, prefix)
  write(out, byte_u(value, 24))
  write(out, byte_u(value, 16))
  write(out, byte_u(value, 8))
  write(out, byte_u(value, 0))
end

local function write_u64(out, prefix, value)
  write(out, prefix)
  write(out, byte_u(value, 56))
  write(out, byte_u(value, 48))
  write(out, byte_u(value, 40))
  write(out, byte_u(value, 32))
  write(out, byte_u(value, 24))
  write(out, byte_u(value, 16))
  write(out, byte_u(value, 8))
  write(out, byte_u(value, 0))
end

------------------------------------------------------------------------------------------------

local function write_array(out, len, value)
  if len <= 0xF then
    write(out, 0x90 + len) -- fixarray
  elseif len <= 0xFFFF then
    write_u16(out, 0xDC, len) -- array16
  elseif len <= 0xFFFFFFFF then
    write_u32(out, 0xDD, len) -- array32
  else
    error("table too long")
  end

  for _, v in pairs(value) do
    write_any(out, v)
  end
end

local function write_map(out, len, value)
  if len <= 0xF then
    write(out, 0x80 + len) -- fixmap
  elseif len <= 0xFFFF then
    write_u16(out, 0xDE, len) -- map16
  elseif len <= 0xFFFFFFFF then
    write_u32(out, 0xDF, len) -- map32
  else
    error("table too long")
  end

  for k, v in pairs(value) do
    write_any(out, k)
    write_any(out, v)
  end
end

------------------------------------------------------------------------------------------------

encoders["nil"] = function (value, out)
  write(out, 0xC0)
end

encoders["boolean"] = function (value, out)
  if value then
    write(out, 0xC3)
  else
    write(out, 0xC2)
  end
end

encoders["number"] = function (value, out)
  if math.floor(value) ~= value then
    error("not float " .. value)
  end

  if value >= 0 then
    if value < 0x80 then 
      write(out, value) -- fixint (positive)
    elseif value <= 0xFF then
      write_u8(out, 0xCC, value) -- uint8
    elseif value <= 0xFFFF then
      write_u16(out, 0xCD, value) -- uint16
    elseif value <= 0xFFFFFFFF then
      write_u32(out, 0xCE, value) -- uint32
    else
      write_u64(out, 0xCF, value) -- uint64
    end
  else
    if value >= -0x20 then
      write(out, 0x100 + value) -- fixint (negative)
    elseif value >= -0x80 then
      write(out, 0xD0) -- int8
      write(out, 0x100 + value) 
    elseif value >= -0x8000 then
      write(out, 0xD1) -- int16
      write(out, byte_i(value, 8))
      write(out, byte_i(value, 0))
    elseif value >= -0x80000000 then
      write(out, 0xD2) -- int32
      write(out, byte_i(value, 24))
      write(out, byte_i(value, 16))
      write(out, byte_i(value, 8))
      write(out, byte_i(value, 0))
    else
      write(out, 0xD2) -- int64
      write(out, byte_i(value, 56))
      write(out, byte_i(value, 48))
      write(out, byte_i(value, 40))
      write(out, byte_i(value, 32))
      write(out, byte_i(value, 24))
      write(out, byte_i(value, 16))
      write(out, byte_i(value, 8))
      write(out, byte_i(value, 0))
    end
  end
end

encoders["string"] = function (value, out)
  local len = #value

  if len <= 0x1F then
    write(out, 0xA0 + len) -- fixstr
  elseif len <= 0xFF then
    write_u8(out, 0xD9, len) -- str8
  elseif len <= 0xFFFF then
    write_u16(out, 0xDA, len) -- str16
  elseif len <= 0xFFFFFFFF then
    write_u32(out, 0xDB, len) -- str32
  else
    error("string too long")
  end

  for i = 1, len do
    write(out, string.byte(value, i))
  end
end

encoders["table"] = function (value, out)
  local len = 0
  local array = true

  for k, v in pairs(value) do
    len = len + 1
    array = array and (k == len)
  end

  if array then
    write_array(out, len, value)
  else
    write_map(out, len, value)
  end
end

------------------------------------------------------------------------------------------------

MessagePack = MessagePack or {}

function MessagePack.encode_u8(value)
  local result_u8 = {}
  
  write_any(result_u8, value)

  print("LEN " .. #(result_u8))
  local dump = ""
  for i = 1, #result_u8 do
    dump = dump .. string.format("%02x ", result_u8[i])
  end
  print(dump)

  return result_u8
end

function MessagePack.encode_i32(value)
  local result_u8 = MessagePack.encode_u8(value)
  local result_i32 = {}

  for i = 1, #result_u8, 4 do
    local b0 = result_u8[i + 0]
    local b1 = result_u8[i + 1] or 0
    local b2 = result_u8[i + 2] or 0
    local b3 = result_u8[i + 3] or 0
    local word = b0 * 0x1000000 + b1 * 0x10000 + b2 * 0x100 + b3 - 0x80000000
    write(result_i32, word)
  end

  return result_i32
end

------------------------------------------------------------------------------------------------
